==> concurrent_plugin-0.4.1-py3-none-any.whl: concurrent_plugin/infinfs/mount_service.py <==
import socket
import time
import os
import traceback
from typing import Any, Union
from concurrent_plugin.infinfs import infinmount
import json
import yaml
from mlflow.tracking import MlflowClient
import psutil
from concurrent_plugin.concurrent_backend import MOUNT_SERVICE_READY_MARKER_FILE
import logging
import requests
import subprocess
# importing it as kubernetes.client since 'client' is used in the code in some places as the Mlflow client.  this mlflow 'client' conflicts with 'from kubernetes import client'.  Need to cleanup.
import kubernetes.client
from kubernetes import client, config
import dpath
from datetime import datetime

logger = logging.getLogger()
logger.setLevel(logging.INFO)


FUSE_DEBUG_FILE = '/tmp/fuse_debug.log'
VERBOSE = False

def parse_mount_request(data):
    req = json.loads(data.decode('utf-8'))
    if req['use_cache'].lower() == 'false':
        use_cache = False
    else:
        use_cache = True
    if req['shadow_path'].lower() == 'none':
        shadow_path = None
    else:
        shadow_path = req['shadow_path']
    return req['mount_path'], req['mount_spec'], shadow_path, use_cache


def print_info(*args):
    print(str(datetime.utcnow()), *args, flush=True)

def mount_service_ready():
    ##Create empty marker file
    if not os.path.exists(MOUNT_SERVICE_READY_MARKER_FILE):
        with open(MOUNT_SERVICE_READY_MARKER_FILE, "w"):
            pass

def read_token(token_file):
    with open(token_file, 'r') as tfh:
        token_file_content = tfh.read()
        for token_line in token_file_content.splitlines():
            if token_line.startswith('Token='):
                return token_line[6:]
    return None

def launch_dag_controller():
    infinstor_token = read_token('/root/.concurrent/token')
    mlflow_parallels_uri = os.environ['MLFLOW_CONCURRENT_URI']
    dag_execution_id = os.environ['DAG_EXECUTION_ID']
    dagid = os.environ['DAGID']
    periodic_run_name = os.environ.get('PERIODIC_RUN_NAME')
    periodic_run_frequency = os.getenv('PERIODIC_RUN_FREQUENCY')
    periodic_run_start_time = os.getenv('PERIODIC_RUN_START_TIME')
    periodic_run_end_time = os.getenv('PERIODIC_RUN_END_TIME')

    execute_dag_url = mlflow_parallels_uri.rstrip('/') + '/api/2.0/mlflow/parallels/execdag'
    logger.info(execute_dag_url)
    headers = {'Content-Type': 'application/json', 'Authorization': infinstor_token}
    body = {'dagid': dagid, 'dag_execution_id': dag_execution_id, "periodic_run_name": periodic_run_name}
    if periodic_run_frequency:
      body['periodic_run_frequency'] = periodic_run_frequency
    if periodic_run_start_time:
      body['periodic_run_start_time'] = periodic_run_start_time
    if periodic_run_end_time:
      body['periodic_run_end_time'] = periodic_run_end_time
    attempts_left = max_attempts = 3
    while attempts_left > 0:
        attempts_left -= 1
        try:
            response = requests.post(execute_dag_url, json=body, headers = headers)
            logger.info(f"DAG Controller response: {response}")
            return
        except Exception as ex:
            logger.warning(str(ex))
            print(f'Exception in dag controller call, retry {attempts_left} more times')
            if attempts_left > 0:
                ##wait before retrying
                time.sleep(10 * 2 ** (max_attempts - attempts_left))
    else:
        raise Exception("Dag Controller launch failed multiple times")


def upload_logs_for_pod(k8s_client:kubernetes.client.CoreV1Api, run_id, pod_name, pod_namespace, tmp_log_file, container_name):
    try:
        # possible fix if only partial logs are read and full logs can't be read. Note that using follow=True for a running pod may make it wait forever??
        # https://github.com/kubernetes-client/python/issues/199: Passing follow=True to read_namespaced_pod_log makes it never return #199
        pod_logs = k8s_client.read_namespaced_pod_log(pod_name, pod_namespace, container=container_name)
        with open(tmp_log_file, "w") as fh:
            fh.write(pod_logs)
    except Exception as ex:
        logger.warning("Failed to fetch logs for {}, {}: {}".format(run_id, pod_name, ex))
        return

    try:
        client = MlflowClient()
        client.log_artifact(run_id, tmp_log_file, artifact_path='.concurrent/logs')
    except Exception as ex:
        logger.warning("Failed upload logs for {}, {}: {}".format(run_id, pod_name, ex))


def update_mlflow_run(run_id, status):
    client = MlflowClient()
    client.set_terminated(run_id, status)

def _filter_empty_in_dict_list_scalar(dict_list_scalar:Union[list, dict, Any]):
    try:
        # depth first traveral
        if isinstance(dict_list_scalar, dict):
            keys_to_del:list = []
            for k in dict_list_scalar.keys():  
                _filter_empty_in_dict_list_scalar(dict_list_scalar[k])
                
                # check if the 'key' is now None or empty.  If so, remove the 'key'
                if not dict_list_scalar[k]: 
                    # RuntimeError: dictionary changed size during iteration
                    # dict_list_scalar.pop(k)
                    keys_to_del.append(k)
            
            # now delete the keys from the map
            for k in keys_to_del:
                dict_list_scalar.pop(k)
        elif isinstance(dict_list_scalar, list):
            i = 0; length = len(dict_list_scalar)
            while i < length: 
                _filter_empty_in_dict_list_scalar(dict_list_scalar[i])
            
                # check if element is now None or empty.  If so, remove the element from the list
                if not dict_list_scalar[i]:
                    dict_list_scalar.remove(dict_list_scalar[i])
                    i -= 1; length -= 1
                
                i += 1
        else: # this must be a non container, like int, str, datatime.datetime
            pass
    except Exception as e:
        # some excpetion, just log it..
        print(f"_filter_empty_in_dict_list_scalar(): Caught exception: {e}")
        traceback.print_exc()

def log_describe_pod(k8s_client:kubernetes.client.CoreV1Api, run_id, pod_name, pod_namespace, pod_info:kubernetes.client.V1Pod):
    describe_file = "/tmp/describe-" + pod_name + ".txt"
    try:
        events:kubernetes.client.V1EventList = k8s_client.list_namespaced_event(pod_namespace, field_selector=f'involvedObject.name={pod_name}')
        with open(describe_file, "w") as fh:
            pod_info_dict:dict = pod_info.to_dict()
            # remove metadata/managed_fields key
            if pod_info_dict.get('metadata') and pod_info_dict.get('metadata').get('managed_fields'): pod_info_dict['metadata'].pop('managed_fields')            
            _filter_empty_in_dict_list_scalar(pod_info_dict)
            fh.write(yaml.safe_dump(pod_info_dict))
            
            events_dict:dict = events.to_dict()
            # api_version: v1
            # kind: EventList
            # metadata:
            #   resource_version: '292715'
            # items:            
            # - count: 1
            #   first_timestamp: 2023-02-22 05:17:20+00:00
            #   involved_object:
            #       api_version: v1
            #       field_path: spec.containers{sidecar-35-16770430018230000000003}
            #       kind: Pod
            #       name: bird-species-2023-02-22-05-17-19-492406-692lg
            #       namespace: parallelsns
            #       resource_version: '106734654'
            #       uid: 008a2e12-b37c-44ed-963f-c9104ccadb4b
            #   last_timestamp: 2023-02-22 05:17:20+00:00
            #   message: Created container sidecar-35-16770430018230000000003
            #   metadata:
            #       creation_timestamp: 2023-02-22 05:17:20+00:00
            #       managed_fields:
            #       - api_version: v1
            #       fields_type: FieldsV1
            #       manager: kubelet
            #       operation: Update
            #       time: 2023-02-22 05:17:20+00:00
            #       name: bird-species-2023-02-22-05-17-19-492406-692lg.17460dc286d42b11
            #       namespace: parallelsns
            #       resource_version: '292711'
            #       uid: 0d95e5e0-85a5-4bf7-8616-dc4eb586fb64
            #   reason: Created
            #   source:
            #       component: kubelet
            #       host: gke-isstage23-cluster-pool-2-vcpu-8gb-43f4db71-x4r5
            #   type: Normal
            # 
            # remove unwanted fields from the eventList
            for path_to_del in '/metadata/managed_fields','/items/*/metadata', '/items/*/involved_object':
                try:
                    # Given a obj, delete all elements that match the glob.  Returns the number of deleted objects. Raises PathNotFound if no paths are found to delete.
                    dpath.delete(events_dict, path_to_del)
                except dpath.PathNotFound as e:
                    print(f"log_describe_pod(): path_to_del={path_to_del}; exception={e}")
            _filter_empty_in_dict_list_scalar(events_dict)
            fh.write(yaml.safe_dump(events_dict))
            
        client = MlflowClient()
        client.log_artifact(run_id, describe_file, artifact_path='.concurrent/logs')
    except Exception as ex:
        logger.warning('Failed to log describe pod, try again later: ' + str(ex))
        return

def fetch_upload_pod_status_logs(k8s_client:client.CoreV1Api, run_id, pod_name, pod_namespace):
    try:
        pod_info:client.V1Pod = k8s_client.read_namespaced_pod(pod_name, pod_namespace)
        if pod_info.spec.containers[1].name.startswith('sidecar-'):
            sidecar_index = 1
            task_index = 0
        else:
            sidecar_index = 0
            task_index = 1
        task_container_name = pod_info.spec.containers[task_index].name
        side_car_container_name = pod_info.spec.containers[sidecar_index].name
        log_describe_pod(k8s_client, run_id, pod_name, pod_namespace, pod_info)
        upload_logs_for_pod(k8s_client, run_id, pod_name, pod_namespace, "/tmp/run-logs.txt",
                            container_name=task_container_name)
        upload_logs_for_pod(k8s_client, run_id, pod_name, pod_namespace, f"/tmp/sidecar-logs.txt",
                            container_name=side_car_container_name)
        task_container_status = pod_info.status
        container_statuses = task_container_status.container_statuses
        task = container_statuses[task_index]
        task_container_state = task.state
        if task_container_state.running:
            print(f"Task container is in running state. Continuing to loop")
            return True
        elif task_container_state.terminated:
            print(f"Task container is in terminated state. Exiting loop")
            return False
        elif task_container_state.waiting:
            print(f"Task container is in waiting state. Continuing to loop")
            return True
        else:
            print(f"Task container is in unknown state. Continuing to loop")
            return True
    except Exception as ex:
        print(f"fetch_upload_pod_status_logs: caught {ex}. Continuing to loop")
        return True # continue looping

def get_task_exit_code(k8s_client, pod_name, pod_namespace, num_attempt=1):
    max_attempts = 3
    pod_info = k8s_client.read_namespaced_pod(pod_name, pod_namespace)
    try:
        if pod_info.spec.containers[1].name.startswith('sidecar-'):
            exitCode = pod_info.status.container_statuses[0].state.terminated.exit_code
        else:
            exitCode = pod_info.status.container_statuses[1].state.terminated.exit_code
        logger.info("Task container finished with exitCode " + str(exitCode))
        return exitCode
    except Exception as ex:
        logger.error("Exception in getting exit code " + str(ex))
        logger.error("Trying {0} more time(s)".format(max_attempts-num_attempt))
        if num_attempt < max_attempts:
            time.sleep(5* 2**num_attempt)
            return get_task_exit_code(k8s_client, pod_name, pod_namespace, num_attempt=num_attempt+1)
        else:
            return -1


if __name__ == '__main__':
    print_info("Starting..")
    HOST = "127.0.0.1"
    PORT = 7963
    last_upload_time = time.time()
    start_time = time.time()
    print("Environment #", os.environ)
    config.load_incluster_config()
    k8s_client:client.CoreV1Api = client.CoreV1Api()
    run_id = os.getenv('MLFLOW_RUN_ID')
    pod_name = os.getenv('MY_POD_NAME')
    pod_namespace = os.getenv('MY_POD_NAMESPACE')
    dag_execution_id = os.getenv('DAG_EXECUTION_ID')
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(15)
        s.bind((HOST, PORT))
        print('Mount/Monitor service starting for runid {0}, and podname {1}'.format(run_id, pod_name), flush=True)
        print('Listening on port {}:{}'.format(HOST, PORT), flush=True)
        s.listen()
        mount_service_ready()
        while True:
            print_info('Waiting for request..')
            try:
                conn, addr = s.accept()
            except socket.timeout:
                print_info('accept timed out')
                pass
            else:
                with conn:
                    print_info(f"Connected by {addr}")
                    data = conn.recv(1024*16)
                    if not data:
                        time.sleep(1)
                        continue
                    try:
                        mount_path, mount_spec, shadow_path, use_cache = parse_mount_request(data)
                        print_info("mount request {}, {}, {}, {}".format(
                            mount_path, mount_spec, shadow_path, use_cache))
                        infinmount.perform_mount(mount_path, mount_spec, shadow_path=shadow_path, use_cache=use_cache)
                        response = "success".encode('utf-8')
                        print_info("mount successful")
                    except Exception as ex:
                        print_info('Exception in mounting: '+str(ex))
                        response = str(ex).encode('utf-8')
                    conn.send(response)
            if not fetch_upload_pod_status_logs(k8s_client, run_id, pod_name, pod_namespace):
                print_info("Task process done, exiting mount service")
                exitCode = get_task_exit_code(k8s_client, pod_name, pod_namespace)
                if exitCode == 0:
                    update_mlflow_run(run_id, "FINISHED")
                else:
                    update_mlflow_run(run_id, "FAILED")
                fetch_upload_pod_status_logs(k8s_client, run_id, pod_name, pod_namespace)
                if dag_execution_id:
                    launch_dag_controller()
                else:
                    print_info('Not a dag execution, skip dag controller')
                exit(0)

==> concurrent_plugin-0.4.2-py3-none-any.whl: concurrent_plugin/infinfs/mount_service.py <==
import socket
import time
import os
import traceback
from typing import Any, Union
from concurrent_plugin.infinfs import infinmount
import json
import yaml
from mlflow.tracking import MlflowClient
import psutil
from concurrent_plugin.concurrent_backend import MOUNT_SERVICE_READY_MARKER_FILE
import logging
import requests
import subprocess
# importing it as kubernetes.client since 'client' is used in the code in some places as the Mlflow client.  this mlflow 'client' conflicts with 'from kubernetes import client'.  Need to cleanup.
import kubernetes.client
from kubernetes import client, config
import dpath
from datetime import datetime

logger = logging.getLogger()
logger.setLevel(logging.INFO)


FUSE_DEBUG_FILE = '/tmp/fuse_debug.log'
VERBOSE = False

def parse_mount_request(data):
    req = json.loads(data.decode('utf-8'))
    if req['use_cache'].lower() == 'false':
        use_cache = False
    else:
        use_cache = True
    if req['shadow_path'].lower() == 'none':
        shadow_path = None
    else:
        shadow_path = req['shadow_path']
    return req['mount_path'], req['mount_spec'], shadow_path, use_cache


def print_info(*args):
    print(str(datetime.utcnow()), *args, flush=True)

def mount_service_ready():
    ##Create empty marker file
    if not os.path.exists(MOUNT_SERVICE_READY_MARKER_FILE):
        with open(MOUNT_SERVICE_READY_MARKER_FILE, "w"):
            pass

def read_token(token_file):
    with open(token_file, 'r') as tfh:
        token_file_content = tfh.read()
        for token_line in token_file_content.splitlines():
            if token_line.startswith('Token='):
                return token_line[6:]
    return None

def launch_dag_controller():
    infinstor_token = read_token('/root/.concurrent/token')
    mlflow_parallels_uri = os.environ['MLFLOW_CONCURRENT_URI']
    dag_execution_id = os.environ['DAG_EXECUTION_ID']
    dagid = os.environ['DAGID']
    periodic_run_name = os.environ.get('PERIODIC_RUN_NAME')
    periodic_run_frequency = os.getenv('PERIODIC_RUN_FREQUENCY')
    periodic_run_start_time = os.getenv('PERIODIC_RUN_START_TIME')
    periodic_run_end_time = os.getenv('PERIODIC_RUN_END_TIME')

    execute_dag_url = mlflow_parallels_uri.rstrip('/') + '/api/2.0/mlflow/parallels/execdag'
    logger.info(execute_dag_url)
    headers = {'Content-Type': 'application/json', 'Authorization': infinstor_token}
    body = {'dagid': dagid, 'dag_execution_id': dag_execution_id, "periodic_run_name": periodic_run_name}
    if periodic_run_frequency:
      body['periodic_run_frequency'] = periodic_run_frequency
    if periodic_run_start_time:
      body['periodic_run_start_time'] = periodic_run_start_time
    if periodic_run_end_time:
      body['periodic_run_end_time'] = periodic_run_end_time
    attempts_left = max_attempts = 3
    while attempts_left > 0:
        attempts_left -= 1
        try:
            response = requests.post(execute_dag_url, json=body, headers = headers)
            logger.info(f"DAG Controller response: {response}")
            return
        except Exception as ex:
            logger.warning(str(ex))
            print(f'Exception in dag controller call, retry {attempts_left} more times')
            if attempts_left > 0:
                ##wait before retrying
                time.sleep(10 * 2 ** (max_attempts - attempts_left))
    else:
        raise Exception("Dag Controller launch failed multiple times")


def upload_logs_for_pod(k8s_client:kubernetes.client.CoreV1Api, run_id, pod_name, pod_namespace, tmp_log_file, container_name):
    try:
        # possible fix if only partial logs are read and full logs can't be read. Note that using follow=True for a running pod may make it wait forever??
        # https://github.com/kubernetes-client/python/issues/199: Passing follow=True to read_namespaced_pod_log makes it never return #199
        pod_logs = k8s_client.read_namespaced_pod_log(pod_name, pod_namespace, container=container_name)
        with open(tmp_log_file, "w") as fh:
            fh.write(pod_logs)
    except Exception as ex:
        logger.warning("Failed to fetch logs for {}, {}: {}".format(run_id, pod_name, ex))
        return

    try:
        client = MlflowClient()
        client.log_artifact(run_id, tmp_log_file, artifact_path='.concurrent/logs')
    except Exception as ex:
        logger.warning("Failed upload logs for {}, {}: {}".format(run_id, pod_name, ex))


def update_mlflow_run(run_id, status):
    client = MlflowClient()
    client.set_terminated(run_id, status)

def _filter_empty_in_dict_list_scalar(dict_list_scalar:Union[list, dict, Any]):
    """
    given a 'dict' or 'list' as input, removes all elements in these containers that are empty: scalars with None, strings that are '', lists and dicts that are empty.  Note that the filtering is in-place: modifies the passed list or dict

    Args:
        dict_list_scalar (Union[list, dict, Any]): see above
    """
    try:
        # depth first traveral
        if isinstance(dict_list_scalar, dict):
            keys_to_del:list = []
            for k in dict_list_scalar.keys():  
                _filter_empty_in_dict_list_scalar(dict_list_scalar[k])
                
                # check if the 'key' is now None or empty.  If so, remove the 'key'
                if not dict_list_scalar[k]: 
                    # cannont do dict.pop(): RuntimeError: dictionary changed size during iteration
                    # dict_list_scalar.pop(k)
                    keys_to_del.append(k)
            
            # now delete the keys from the map
            for k in keys_to_del:
                dict_list_scalar.pop(k)
        elif isinstance(dict_list_scalar, list):
            i = 0; length = len(dict_list_scalar)
            while i < length: 
                _filter_empty_in_dict_list_scalar(dict_list_scalar[i])
            
                # check if element is now None (if scalar) or empty (if list or dict).  If so, remove the element from the list
                if not dict_list_scalar[i]:
                    dict_list_scalar.remove(dict_list_scalar[i])
                    i -= 1; length -= 1
                
                i += 1
        else: # this must be a non container, like int, str, datatime.datetime
            pass
    except Exception as e:
        # some excpetion, just log it..
        print(f"_filter_empty_in_dict_list_scalar(): Caught exception: {e}")
        traceback.print_exc()

def log_describe_pod(k8s_client:kubernetes.client.CoreV1Api, run_id, pod_name, pod_namespace, pod_info:kubernetes.client.V1Pod):
    describe_file = "/tmp/describe-" + pod_name + ".txt"
    try:
        events:kubernetes.client.V1EventList = k8s_client.list_namespaced_event(pod_namespace, field_selector=f'involvedObject.name={pod_name}')
        with open(describe_file, "w") as fh:
            pod_info_dict:dict = pod_info.to_dict()
            # remove metadata/managed_fields key
            if pod_info_dict.get('metadata') and pod_info_dict.get('metadata').get('managed_fields'): pod_info_dict['metadata'].pop('managed_fields')            
            _filter_empty_in_dict_list_scalar(pod_info_dict)
            fh.write(yaml.safe_dump(pod_info_dict))
            
            events_dict:dict = events.to_dict()
            # api_version: v1
            # kind: EventList
            # metadata:
            #   resource_version: '292715'
            # items:            
            # - count: 1
            #   first_timestamp: 2023-02-22 05:17:20+00:00
            #   involved_object:
            #       api_version: v1
            #       field_path: spec.containers{sidecar-35-16770430018230000000003}
            #       kind: Pod
            #       name: bird-species-2023-02-22-05-17-19-492406-692lg
            #       namespace: parallelsns
            #       resource_version: '106734654'
            #       uid: 008a2e12-b37c-44ed-963f-c9104ccadb4b
            #   last_timestamp: 2023-02-22 05:17:20+00:00
            #   message: Created container sidecar-35-16770430018230000000003
            #   metadata:
            #       creation_timestamp: 2023-02-22 05:17:20+00:00
            #       managed_fields:
            #       - api_version: v1
            #       fields_type: FieldsV1
            #       manager: kubelet
            #       operation: Update
            #       time: 2023-02-22 05:17:20+00:00
            #       name: bird-species-2023-02-22-05-17-19-492406-692lg.17460dc286d42b11
            #       namespace: parallelsns
            #       resource_version: '292711'
            #       uid: 0d95e5e0-85a5-4bf7-8616-dc4eb586fb64
            #   reason: Created
            #   source:
            #       component: kubelet
            #       host: gke-isstage23-cluster-pool-2-vcpu-8gb-43f4db71-x4r5
            #   type: Normal
            # 
            # remove unwanted fields from the eventList
            for path_to_del in '/metadata/managed_fields','/items/*/metadata', '/items/*/involved_object':
                try:
                    # Given a obj, delete all elements that match the glob.  Returns the number of deleted objects. Raises PathNotFound if no paths are found to delete.
                    dpath.delete(events_dict, path_to_del)
                except dpath.PathNotFound as e:
                    print(f"log_describe_pod(): path_to_del={path_to_del}; exception={e}")
            _filter_empty_in_dict_list_scalar(events_dict)
            fh.write(yaml.safe_dump(events_dict))
            
        client = MlflowClient()
        client.log_artifact(run_id, describe_file, artifact_path='.concurrent/logs')
    except Exception as ex:
        logger.warning('Failed to log describe pod, try again later: ' + str(ex))
        return

def fetch_upload_pod_status_logs(k8s_client:client.CoreV1Api, run_id, pod_name, pod_namespace):
    try:
        pod_info:client.V1Pod = k8s_client.read_namespaced_pod(pod_name, pod_namespace)
        if pod_info.spec.containers[1].name.startswith('sidecar-'):
            sidecar_index = 1
            task_index = 0
        else:
            sidecar_index = 0
            task_index = 1
        task_container_name = pod_info.spec.containers[task_index].name
        side_car_container_name = pod_info.spec.containers[sidecar_index].name
        log_describe_pod(k8s_client, run_id, pod_name, pod_namespace, pod_info)
        upload_logs_for_pod(k8s_client, run_id, pod_name, pod_namespace, "/tmp/run-logs.txt",
                            container_name=task_container_name)
        upload_logs_for_pod(k8s_client, run_id, pod_name, pod_namespace, f"/tmp/sidecar-logs.txt",
                            container_name=side_car_container_name)
        task_container_status = pod_info.status
        container_statuses = task_container_status.container_statuses
        task = container_statuses[task_index]
        task_container_state = task.state
        if task_container_state.running:
            print(f"Task container is in running state. Continuing to loop")
            return True
        elif task_container_state.terminated:
            print(f"Task container is in terminated state. Exiting loop")
            return False
        elif task_container_state.waiting:
            print(f"Task container is in waiting state. Continuing to loop")
            return True
        else:
            print(f"Task container is in unknown state. Continuing to loop")
            return True
    except Exception as ex:
        print(f"fetch_upload_pod_status_logs: caught {ex}. Continuing to loop")
        return True # continue looping

def get_task_exit_code(k8s_client, pod_name, pod_namespace, num_attempt=1):
    max_attempts = 3
    pod_info = k8s_client.read_namespaced_pod(pod_name, pod_namespace)
    try:
        if pod_info.spec.containers[1].name.startswith('sidecar-'):
            exitCode = pod_info.status.container_statuses[0].state.terminated.exit_code
        else:
            exitCode = pod_info.status.container_statuses[1].state.terminated.exit_code
        logger.info("Task container finished with exitCode " + str(exitCode))
        return exitCode
    except Exception as ex:
        logger.error("Exception in getting exit code " + str(ex))
        logger.error("Trying {0} more time(s)".format(max_attempts-num_attempt))
        if num_attempt < max_attempts:
            time.sleep(5* 2**num_attempt)
            return get_task_exit_code(k8s_client, pod_name, pod_namespace, num_attempt=num_attempt+1)
        else:
            return -1


if __name__ == '__main__':
    print_info("Starting..")
    HOST = "127.0.0.1"
    PORT = 7963
    last_upload_time = time.time()
    start_time = time.time()
    print("Environment #", os.environ)
    config.load_incluster_config()
    k8s_client:client.CoreV1Api = client.CoreV1Api()
    run_id = os.getenv('MLFLOW_RUN_ID')
    pod_name = os.getenv('MY_POD_NAME')
    pod_namespace = os.getenv('MY_POD_NAMESPACE')
    dag_execution_id = os.getenv('DAG_EXECUTION_ID')
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(15)
        s.bind((HOST, PORT))
        print('Mount/Monitor service starting for runid {0}, and podname {1}'.format(run_id, pod_name), flush=True)
        print('Listening on port {}:{}'.format(HOST, PORT), flush=True)
        s.listen()
        mount_service_ready()
        while True:
            print_info('Waiting for request..')
            try:
                conn, addr = s.accept()
            except socket.timeout:
                print_info('accept timed out')
                pass
            else:
                with conn:
                    print_info(f"Connected by {addr}")
                    data = conn.recv(1024*16)
                    if not data:
                        time.sleep(1)
                        continue
                    try:
                        mount_path, mount_spec, shadow_path, use_cache = parse_mount_request(data)
                        print_info("mount request {}, {}, {}, {}".format(
                            mount_path, mount_spec, shadow_path, use_cache))
                        infinmount.perform_mount(mount_path, mount_spec, shadow_path=shadow_path, use_cache=use_cache)
                        response = "success".encode('utf-8')
                        print_info("mount successful")
                    except Exception as ex:
                        print_info('Exception in mounting: '+str(ex))
                        response = str(ex).encode('utf-8')
                    conn.send(response)
            if not fetch_upload_pod_status_logs(k8s_client, run_id, pod_name, pod_namespace):
                print_info("Task process done, exiting mount service")
                exitCode = get_task_exit_code(k8s_client, pod_name, pod_namespace)
                if exitCode == 0:
                    update_mlflow_run(run_id, "FINISHED")
                else:
                    update_mlflow_run(run_id, "FAILED")
                fetch_upload_pod_status_logs(k8s_client, run_id, pod_name, pod_namespace)
                if dag_execution_id:
                    launch_dag_controller()
                else:
                    print_info('Not a dag execution, skip dag controller')
                exit(0)
